import asyncio
import os
import shutil
import sys
from functools import lru_cache
from pathlib import Path
from typing import Awaitable, Optional

import yaml
from annet.annlib.lib import (  # pylint: disable=unused-import
    ContextOrderedDict,
    HuaweiNumBlock,
    LMSMatcher,
    add_annotation,
    catch_ctrl_c,
    cisco_collapse_vlandb,
    cisco_expand_vlandb,
    find_exc_in_stack,
    find_modules,
    first,
    flatten,
    huawei_collapse_vlandb,
    huawei_expand_vlandb,
    huawei_iface_ranges,
    is_relative,
    jinja_render,
    jun_activate,
    jun_is_inactive,
    juniper_fmt_prefix_lists_acl,
    juniper_port_split,
    make_ip4_mask,
    mako_render,
    merge_dicts,
    percentile,
    uniq,
)
from contextlog import get_logger


_TEMPLATE_CONTEXT_PATH: Optional[str] = None
_DEFAULT_CONTEXT_PATH: Optional[str] = None


def get_template_context_path() -> str:
    if _TEMPLATE_CONTEXT_PATH is None:
        set_template_context_path(str(Path(sys.modules["annet"].__file__).parent / "configs/context.yml"))
    return _TEMPLATE_CONTEXT_PATH


def set_template_context_path(path: str) -> None:
    global _TEMPLATE_CONTEXT_PATH  # pylint: disable=global-statement
    _TEMPLATE_CONTEXT_PATH = path


def get_default_context_path() -> str:
    if _DEFAULT_CONTEXT_PATH is None:
        set_default_context_path("~/.annet/context.yml")
    return _DEFAULT_CONTEXT_PATH


def set_default_context_path(path: str) -> None:
    global _DEFAULT_CONTEXT_PATH  # pylint: disable=global-statement
    _DEFAULT_CONTEXT_PATH = path


@lru_cache(maxsize=1)
def _get_template_context():
    with open(get_template_context_path()) as f:
        return yaml.safe_load(f)


def get_context_path(touch: Optional[bool] = False) -> str:
    path = Path(os.getenv("ANN_CONTEXT_CONFIG_PATH", get_default_context_path())).expanduser().absolute()
    if not path.exists():
        src = get_template_context_path()
        if not touch:
            return str(src)
        try:
            # populate path with default configuration
            path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(src, path)
        except shutil.SameFileError:
            pass
    return str(path)


@lru_cache(maxsize=1)
def get_context() -> dict:
    with open(get_context_path()) as f:
        raw = yaml.safe_load(f)
    _fill_in_default_generator_modules(raw)
    context_name = os.getenv("ANN_SELECTED_CONTEXT", raw["selected_context"])
    res = {k: raw[k][v] for k, v in raw["context"][context_name].items()}
    if "ANN_GENERATORS_CONTEXT" in os.environ:  # an undocumented hack to maintain backwards compatibility; TODO: remove
        res["generators"] = raw["generators"][os.getenv("ANN_GENERATORS_CONTEXT")]
    return res


@lru_cache(maxsize=1)
def _warn_no_generators_in_context():
    get_logger().warning(
        "Older version of the context configuration found. Getting generators references from the template context"
    )


def _fill_in_default_generator_modules(raw: dict) -> bool:
    """Backwards compatibility hack to add existing generators refs to context"""
    if "generators" not in raw:
        _warn_no_generators_in_context()
        raw["generators"] = _get_template_context()["generators"]
        for dst_context in raw["context"].values():
            dst_context["generators"] = "default"
        return True
    return False


def repair_context_file() -> None:
    path = get_context_path()
    with open(path) as f:
        data = yaml.safe_load(f)
    if _fill_in_default_generator_modules(data):
        with open(path, "w") as f:
            yaml.dump(data, f, sort_keys=False)


def do_async(coro: Awaitable):
    loop = asyncio.get_event_loop()
    res = loop.run_until_complete(coro)
    return res



import functools
from collections import OrderedDict as odict
from collections import namedtuple

from annet.annlib.rbparser import platform, syntax
from annet.annlib.rbparser.deploying import compile_messages
from valkit.common import valid_bool, valid_number, valid_string_list
from valkit.python import valid_object_path

from .common import import_rulebook_function


Answer = namedtuple("Answer", ("text", "send_nl"))
DEFAULT_TIMEOUT = 30
DEFAULT_SEND_NL = True
DEFAULT_APPLY_LOGIC = "common.apply"


# =====
@functools.lru_cache()
def compile_deploying_text(text, vendor):
    return _compile_deploying(
        tree=syntax.parse_text(text, params_scheme={
            "timeout": {
                "validator": lambda arg: valid_number(arg, min=1, type=float),
                "default": 30,
            },
            "send_nl": {
                "validator": valid_bool,
                "default": True,
            },
            "apply_logic": {
                "validator": valid_object_path,
                "default": DEFAULT_APPLY_LOGIC,
            },
            "ifcontext": {
                "validator": valid_string_list,
                "default": [],
            }
        }),
        reverse_prefix=platform.VENDOR_REVERSES[vendor],
    )


# =====
def _compile_deploying(tree, reverse_prefix):
    deploying = odict()
    for (rule_id, attrs) in tree.items():
        if attrs["type"] == "normal" and not attrs["row"].startswith(("ignore:", "dialog:")):
            (ignore, dialogs) = compile_messages(attrs["children"])
            deploying[rule_id] = {
                "attrs": {
                    "regexp": syntax.compile_row_regexp(attrs["row"]),
                    "timeout": attrs["params"]["timeout"],
                    "apply_logic": import_rulebook_function(attrs["params"]["apply_logic"]),
                    "ignore": ignore,
                    "dialogs": dialogs,
                    "ifcontext": attrs["params"]["ifcontext"],
                },
                "children": _compile_deploying(attrs["children"], reverse_prefix),
            }
    return deploying


def match_deploy_rule(rules, cmd_path, context):
    for (depth, row) in enumerate(cmd_path):
        for rule in rules.values():
            if rule["attrs"]["regexp"].match(row):
                ifcontext = rule["attrs"]["ifcontext"]
                if syntax.match_context(ifcontext, context):
                    if depth == len(cmd_path) - 1:
                        return rule
                    else:
                        rules = rule["children"]
                        if len(rules) == 0:
                            break
    # default match
    return {
        "attrs": {
            "regexp": syntax.compile_row_regexp("~"),
            "timeout": DEFAULT_TIMEOUT,
            "apply_logic": import_rulebook_function(DEFAULT_APPLY_LOGIC),
            "ignore": [],
            "dialogs": odict(),
        },
        "children": odict(),
    }



from logging import getLogger
from typing import Optional, List, Union

from annetbox.v24 import models as api_models
from annetbox.v24.client_sync import NetboxV24

from annet.adapters.netbox.common import models
from annet.adapters.netbox.common.manufacturer import (
    is_supported, get_hw, get_breed,
)
from annet.adapters.netbox.common.query import NetboxQuery
from annet.adapters.netbox.common.storage_opts import NetboxStorageOpts
from annet.storage import Storage

logger = getLogger(__name__)


def extend_device_ip(
        ip: Optional[api_models.DeviceIp],
) -> Optional[models.DeviceIp]:
    if not ip:
        return None
    return models.DeviceIp(
        address=ip.address,
        id=ip.id,
        display=ip.address,
        family=ip.family,
    )


def extend_label(
        label: Optional[api_models.Label],
) -> Optional[models.Label]:
    if not label:
        return None
    return models.Label(
        label=label.label,
        value=str(label.value),
    )


def extend_device(
        device: api_models.Device, storage: Storage,
) -> models.NetboxDevice:
    manufacturer = device.device_type.manufacturer.name
    model = device.device_type.model
    platform_name = ""
    if device.platform:
        platform_name = device.platform.name

    return models.NetboxDevice(
        url=device.url,
        id=device.id,
        name=device.name,
        display=device.display_name,
        device_type=device.device_type,
        device_role=device.device_role,
        tenant=device.tenant,
        platform=device.platform,
        serial=device.serial,
        asset_tag=device.asset_tag,
        site=device.site,
        rack=device.rack,
        position=device.position,
        face=extend_label(device.face),
        status=device.status,
        primary_ip=extend_device_ip(device.primary_ip),
        primary_ip4=extend_device_ip(device.primary_ip4),
        primary_ip6=extend_device_ip(device.primary_ip6),
        tags=[models.Entity(0, tag) for tag in device.tags],
        custom_fields=device.custom_fields,  # ???
        created=device.created,
        last_updated=device.last_updated,

        fqdn=device.name,
        hostname=device.name,
        hw=get_hw(manufacturer, model, platform_name),
        breed=get_breed(manufacturer, model),
        interfaces=[],
        neighbours_ids=[],
        storage=storage,
    )


def extend_interface(interface: api_models.Interface) -> models.Interface:
    return models.Interface(
        id=interface.id,
        name=interface.name,
        device=interface.device,
        enabled=interface.enabled,
        display=interface.name,
        ip_addresses=[],
    )


def extend_ip(ip: api_models.IpAddress) -> models.IpAddress:
    return models.IpAddress(
        id=ip.id,
        assigned_object_id=ip.interface.id,
        display=ip.address,
        family=models.IpFamily(
            value=ip.family,
            label=str(ip.family),
        ),
        address=ip.address,
        status=extend_label(ip.status),
        tags=[models.Entity(0, tag) for tag in ip.tags],
        created=ip.created,
        last_updated=ip.last_updated,
    )


class NetboxStorageV24(Storage):
    def __init__(self, opts: Optional[NetboxStorageOpts] = None):
        self.netbox = NetboxV24(
            url=opts.url,
            token=opts.token,
        )

    def __enter__(self):
        return self

    def __exit__(self, _, __, ___):
        pass

    def resolve_object_ids_by_query(self, query: NetboxQuery):
        return [
            d.id for d in self._load_devices(query)
        ]

    def resolve_fdnds_by_query(self, query: NetboxQuery):
        return [
            d.name for d in self._load_devices(query)
        ]

    def make_devices(
            self,
            query: Union[NetboxQuery, list],
            preload_neighbors=False,
            use_mesh=None,
            preload_extra_fields=False,
            **kwargs,
    ) -> List[models.NetboxDevice]:
        if isinstance(query, list):
            query = NetboxQuery.new(query)
        device_ids = {
            device.id: extend_device(device=device, storage=self)
            for device in self._load_devices(query)
        }
        if not device_ids:
            return []

        interfaces = self._load_interfaces(list(device_ids))
        for interface in interfaces:
            device_ids[interface.device.id].interfaces.append(interface)
        return list(device_ids.values())

    def _load_devices(self, query: NetboxQuery) -> List[api_models.Device]:
        if not query.globs:
            return []
        return [
            device
            for device in self.netbox.dcim_all_devices().results
            if _match_query(query, device)
            if is_supported(device.device_type.manufacturer.name)
        ]

    def _load_interfaces(self, device_ids: List[int]) -> List[
        models.Interface]:
        interfaces = self.netbox.dcim_all_interfaces(device_id=device_ids)
        extended_ifaces = {
            interface.id: extend_interface(interface)
            for interface in interfaces.results
        }

        ips = self.netbox.ipam_all_ip_addresses(interface_id=list(extended_ifaces))
        for ip in ips.results:
            extended_ip = extend_ip(ip)
            interface = extended_ifaces[extended_ip.assigned_object_id]
            interface.ip_addresses.append(extended_ip)
        return list(extended_ifaces.values())

    def get_device(
            self, obj_id, preload_neighbors=False, use_mesh=None,
            **kwargs,
    ) -> models.NetboxDevice:
        device = self.netbox.dcim_device(obj_id)
        res = extend_device(device=device, storage=self)
        res.interfaces = self._load_interfaces([device.id])
        return res

    def flush_perf(self):
        pass


def _match_query(query: NetboxQuery, device_data: api_models.Device) -> bool:
    for subquery in query.globs:
        if subquery.strip() in device_data.name:
            return True
    return False




from typing import List

from annet.generators import BaseGenerator, PartialGenerator
from annet.storage import Storage


# ====
class Lldp(PartialGenerator):
    TAGS = ["mgmt", "lldp"]

    def acl_huawei(self, device):
        return """
            lldp
        """

    def run_huawei(self, device):
        yield "lldp enable"

        if device.hw.CE:
            yield "lldp transmit interval 10"

    def acl_nexus(self, device):
        return """
            feature lldp
            lldp
        """

    def run_nexus(self, device):
        yield "feature lldp"
        yield "lldp timer 10"

    def acl_juniper(self, _):
        return """
        protocols    %cant_delete
            lldp
                *
        """

    def run_juniper(self, device):
        with self.multiblock("protocols", "lldp"):
            yield """
                neighbour-port-info-display port-id
                port-description-type interface-alias
                port-id-subtype interface-name
                interface all
            """


def get_generators(store: Storage) -> List[BaseGenerator]:
    return [
        Lldp(store),
    ]



from logging import getLogger

from annet.annlib.netdev.views.hardware import HardwareView

logger = getLogger(__name__)

_VENDORS = {
    "cisco": "Cisco",
    "catalyst": "Cisco Catalyst",
    "nexus": "Cisco Nexus",
    "huawei": "Huawei",
    "juniper": "Juniper",
    "arista": "Arista",
    "pc": "PC",
    "nokia": "Nokia",
    "aruba": "Aruba",
    "routeros": "RouterOS",
    "ribbon": "Ribbon",
}


def get_hw(manufacturer: str, model: str, platform_name: str):
    # by some reason Netbox calls Mellanox SN as MSN, so we fix them here
    if manufacturer == "Mellanox" and model.startswith("MSN"):
        model = model.replace("MSN", "SN", 1)
    vendor = manufacturer + " " + model
    hw = HardwareView(_VENDORS.get(vendor.lower(), vendor), platform_name)
    if not hw:
        raise ValueError(f"unsupported manufacturer {manufacturer}")
    return hw


def get_breed(manufacturer: str, model: str):
    if manufacturer == "Huawei" and model.startswith("CE"):
        return "vrp85"
    elif manufacturer == "Huawei" and model.startswith("NE"):
        return "vrp85"
    elif manufacturer == "Huawei":
        return "vrp55"
    elif manufacturer == "Mellanox":
        return "cuml2"
    elif manufacturer == "Juniper":
        return "jun10"
    elif manufacturer == "Cisco":
        return "ios12"
    elif manufacturer == "Adva":
        return "adva8"
    elif manufacturer == "Arista":
        return "eos4"
    raise ValueError(f"unsupported manufacturer {manufacturer}")


def is_supported(manufacturer: str) -> bool:
    if manufacturer not in (
            "Huawei", "Mellanox", "Juniper", "Cisco", "Adva", "Arista",
    ):
        logger.warning("Unsupported manufacturer `%s`", manufacturer)
        return False
    return True



from annet.annlib.patching import match_row_to_acl
from annet.annlib.rbparser.acl import compile_ref_acl_text


class RefMatchResult:
    def __init__(self, gen_cls=None, groups=None):
        self.elems = {}
        if gen_cls and groups:
            self.elems[gen_cls] = groups

    def gen_cls(self):
        return list(self.elems.keys())

    def groups(self):
        return list(sum(self.elems.values(), []))

    def __iter__(self):
        yield from self.elems.items()

    def __add__(self, other):
        ret = RefMatchResult()
        ret.elems.update(self.elems)
        ret.elems.update(other.elems)
        return ret


class RefMatch:
    def __init__(self, acl, gen_cls):
        self.acl = acl
        self.gen_cls = gen_cls

    def match(self, config):
        ret = RefMatchResult()
        passed = self._match(config, self.acl)
        if passed:
            ret = RefMatchResult(self.gen_cls, passed)
        return ret

    @classmethod
    def _match(cls, config, acl, _path=tuple()):
        ret = []
        for (row, children) in config.items():
            (rule, children_rules) = match_row_to_acl(row, acl)
            if rule and rule["attrs"]["match"]:
                ret.append(rule["attrs"]["match"])
            if rule and children and children_rules:
                ret += cls._match(children, children_rules, _path + (row,))
        return ret


class RefMatcher:
    def __init__(self):
        self.matches = []

    def match(self, config):
        ret = RefMatchResult()
        for rmatch in self.matches:
            ret += rmatch.match(config)
        return ret

    def add(self, acl_text, gen_class):
        acl = compile_ref_acl_text(acl_text)
        self.matches.append(RefMatch(acl, gen_class))


class RefTracker:
    class Root:
        pass

    def __init__(self):
        self.cfgs = {}
        self.mapidx = {}
        self.graph = Graph()
        self.root = self.__class__.Root
        self.map(self.root, self.graph.root_id)

    def map(self, newcls, node_id):
        self.mapidx[node_id] = newcls
        self.mapidx[newcls] = node_id

    def addcls(self, newcls):
        if newcls not in self.mapidx:
            self.map(newcls, self.graph.newnode())
        return self.mapidx[newcls]

    def add(self, refcls, defcls):
        if refcls not in self.mapidx:
            self.add(self.root, refcls)
        ridx = self.addcls(refcls)
        didx = self.addcls(defcls)
        self.graph.connect(ridx, didx)

    def config(self, newcls, config):
        self.cfgs[newcls] = config

    def walk(self):
        ret = []
        for didx, ridx in self.graph.walk():
            dc = self.mapidx[didx]
            rc = self.mapidx[ridx]
            yield dc, rc
        return ret

    def configs(self):
        ret = []
        for dc, rc in self.walk():
            if dc in self.cfgs and rc in self.cfgs:
                ret.append((self.cfgs[dc], self.cfgs[rc]))
        return ret


class Graph:
    def __init__(self):
        self.indices = []
        self.root_id = self.newnode()

    def newnode(self):
        node_id = len(self.indices)
        self.indices.append([0 for _ in range(len(self.indices))])
        for ind in self.indices:
            ind.append(0)
        return node_id

    def connect(self, ridx, didx):
        self.indices[ridx][didx] = 1

    def walk(self):
        def childs(ridx):
            ret = []
            for didx, is_ref in enumerate(self.indices[ridx]):
                if is_ref:
                    ret.append((ridx, didx))
            return ret

        def bfs(run, seen):
            ch = []
            for key in run:
                ridx, didx = key
                if key not in seen:
                    seen.add(key)
                    ch += childs(didx)
                    yield ridx, didx
            if ch:
                yield from bfs(ch, seen)

        for left, right in bfs(childs(self.root_id), set()):
            if left not in (self.root_id, right):
                yield left, right



import functools
import json
import re
from os import path
from typing import Any, Dict

from annet.annlib.netdev.db import find_true_sequences, get_db


@functools.lru_cache(None)
def parse_hw_model(hw_model):
    prepared = _prepare_db()
    (tree, all_sequences) = get_db(prepared)
    true_sequences = find_true_sequences(hw_model, tree)
    return (
        sorted(true_sequences),
        all_sequences.difference(true_sequences),
    )


def _prepare_db() -> Dict[str, Any]:
    try:
        from library.python import resource
        raw = json.loads(resource.resfs_read("contrib/python/annet/annet/annlib/netdev/devdb/data/devdb.json").decode("utf-8"))
    except ImportError:
        with open(path.join(path.dirname(__file__), "data", "devdb.json"), "r") as f:
            raw = json.load(f)
    return {tuple(seq.split(".")): re.compile(regexp) for (seq, regexp) in raw.items()}



class ExecError(Exception):
    """Обработчик этого exception должен залоггировать ошибку и выйти с exit_code 1"""
    pass


class DeployCancelled(Exception):
    """Деплой на устройство был отменен"""
    pass



import os

import colorama

# отключить colorama.init, если стоит env-переменная. Нужно в тестах
if os.environ.get("ANN_FORCE_COLOR", None) not in [None, "", "0", "no"]:
    colorama.init = lambda *_, **__: None



import re
from itertools import groupby
from typing import Generator, List, Mapping, Tuple, Union

from annet.annlib.diff import (  # pylint: disable=unused-import
    colorize_line,
    diff_cmp,
    diff_ops,
    gen_pre_as_diff,
    resort_diff,
)
from annet.annlib.output import format_file_diff

from annet import patching
from annet.cli_args import ShowDiffOptions
from annet.output import output_driver_connector
from annet.storage import Device
from annet.tabparser import make_formatter
from annet.types import Diff, PCDiff


# NOCDEV-1720


def gen_sort_diff(
    diffs: Mapping[Device, Union[Diff, PCDiff]], args: ShowDiffOptions
) -> Generator[Tuple[str, Generator[str, None, None], bool], None, None]:
    """
    Возвращает осортированный дифф, совместимый с write_output
    :param diffs: Маппинг устройства в дифф
    :param args: Параметры коммандной строки
    """
    if args.no_collapse:
        devices_to_diff = {(dev,): diff for dev, diff in diffs.items()}
    else:
        non_pc_diffs = {dev: diff for dev, diff in diffs.items() if not isinstance(diff, PCDiff)}
        devices_to_diff = collapse_diffs(non_pc_diffs)
        devices_to_diff.update({(dev,): diff for dev, diff in diffs.items() if isinstance(diff, PCDiff)})
    for devices, diff_obj in devices_to_diff.items():
        if not diff_obj:
            continue
        if isinstance(diff_obj, PCDiff):
            for diff_file in diff_obj.diff_files:
                diff_text = (
                    "\n".join(diff_file.diff_lines)
                    if args.no_color
                    else "\n".join(format_file_diff(diff_file.diff_lines))
                )
                yield diff_file.label, diff_text, False
        else:
            output_driver = output_driver_connector.get()
            dest_name = ", ".join([output_driver.cfg_file_names(dev)[0] for dev in devices])
            pd = patching.make_pre(resort_diff(diff_obj))
            yield dest_name, gen_pre_as_diff(pd, args.show_rules, args.indent, args.no_color), False


def _transform_text_diff_for_collapsing(text_diff) -> List[str]:
    for line_no, line in enumerate(text_diff):
        text_diff[line_no] = re.sub(r"(snmp-agent .+) cipher \S+ (.+)", r"\1 cipher ENCRYPTED \2", line)
    return text_diff


def _make_text_diff(device: Device, diff: Diff) -> List[str]:
    formatter = make_formatter(device.hw)
    res = formatter.diff(diff)
    return res


def collapse_diffs(diffs: Mapping[Device, Diff]) -> Mapping[Tuple[Device, ...], Diff]:
    """
    Группировка диффов.
    :param diffs:
    :return: дикт аналогичный типу Diff, но с несколькими dev в ключе.
        Нужно учесть что дифы сверяются в отформатированном виде
    """
    diffs_with_test = {dev: [diff, _transform_text_diff_for_collapsing(_make_text_diff(dev, diff))] for dev, diff in
                       diffs.items()}
    res = {}
    for _, collapsed_diff_iter in groupby(sorted(diffs_with_test.items(), key=lambda x: (x[0].hw.vendor, x[1][1])),
                                          key=lambda x: x[1][1]):
        collapsed_diff = list(collapsed_diff_iter)
        res[tuple(x[0] for x in collapsed_diff)] = collapsed_diff[0][1][0]

    return res



VENDOR_REVERSES = {
    "huawei": "undo",
    "cisco": "no",
    "nexus": "no",
    "juniper": "delete",
    "arista": "no",
    "nokia": "delete",
    "routeros": "remove",
    "aruba": "no",
    "pc": "-",
    "ribbon": "delete",
}

VENDOR_DIFF = {
    "huawei": "common.default_diff",
    "cisco": "common.default_diff",
    "nexus": "common.default_diff",
    "juniper": "juniper.default_diff",
    "arista": "common.default_diff",
    "nokia": "juniper.default_diff",
    "routeros": "common.default_diff",
    "aruba": "aruba.default_diff",
    "pc": "common.default_diff",
    "ribbon": "ribbon.default_diff",
}

VENDOR_DIFF_ORDERED = {
    "huawei": "common.ordered_diff",
    "cisco": "common.ordered_diff",
    "nexus": "common.ordered_diff",
    "juniper": "juniper.ordered_diff",
    "arista": "common.ordered_diff",
    "nokia": "juniper.ordered_diff",
    "routeros": "common.ordered_diff",
    "aruba": "common.ordered_diff",
    "pc": "common.ordered_diff",
    "ribbon": "ribbon.default_diff",
}

VENDOR_EXIT = {
    "huawei": "quit",
    "cisco": "exit",
    "nexus": "exit",
    "arista": "exit",
    "juniper": "",
    "nokia": "",
    "routeros": "",
    "aruba": "exit",
    "pc": "",
    "ribbon": "exit",
}
